# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

class AnyMessage(object):
    NONE = 0
    Hello = 1
    Welcome = 2
    Abort = 3
    Challenge = 4
    Authenticate = 5
    Goodbye = 6
    Error = 7
    Publish = 8
    Published = 9
    SubscriberReceived = 10
    Subscribe = 11
    Subscribed = 12
    Unsubscribe = 13
    Unsubscribed = 14
    Event = 15
    EventReceived = 16
    Call = 17
    Cancel = 18
    Result = 19
    Register = 20
    Registered = 21
    Unregister = 22
    Unregistered = 23
    Invocation = 24
    Interrupt = 25
    Yield = 26

